"""Loader class for downloading stock data.

Download stock data from yahoo finance.

"""

from __future__ import annotations

import pickle
from datetime import datetime
from typing import Dict, List, Optional

import pandas as pd
import yfinance as yf

from mypo.market import Market


class Loader(object):
    """Loader class for downloading stock."""

    _tickers: Dict[str, pd.DataFrame]
    _names: Dict[str, str]
    _expense_ratio: Dict[str, float]
    _total_assets: Dict[str, int]
    _daily_volume_10_days: Dict[str, int]

    def __init__(
        self,
        tickers: Optional[Dict[str, pd.DataFrame]] = None,
        names: Optional[Dict[str, str]] = None,
        expense_ratio: Optional[Dict[str, float]] = None,
        total_assets: Optional[Dict[str, int]] = None,
        daily_volume_10_days: Optional[Dict[str, int]] = None,
    ) -> None:  # pragma: no cover
        """Construct this object.

        Args:
            tickers: Tickers.
            names: Names.
            expense_ratio: Expense ratio.
            total_assets: Total assets.
            daily_volume_10_days: Daily volume 10 days.
        """
        self._tickers = {} if tickers is None else tickers
        self._names = {} if names is None else names
        self._expense_ratio = {} if expense_ratio is None else expense_ratio
        self._total_assets = {} if total_assets is None else total_assets
        self._daily_volume_10_days = {} if daily_volume_10_days is None else daily_volume_10_days

    def save(self, filepath: str) -> None:  # pragma: no cover
        """Save market data to file.

        Args:
            filepath: Path to file for storing data.
        """
        with open(filepath, "wb") as bin_file:
            pickle.dump(self, bin_file)

    @staticmethod
    def load(filepath: str) -> Loader:
        """Load market data from file.

        Args:
            filepath: Path to file for loading data.

        Returns:
            Market object
        """
        with open(filepath, "rb") as bin_file:
            value: Loader = pickle.load(bin_file)
            return value

    def get(self, ticker: str, expense_ratio: float = 0.0) -> None:  # pragma: no cover
        """Get stock data of specified ticker.

        Args:
            ticker: Ticker that you want to download stock data.
            expense_ratio: Expense ratio of ticker. The default value is 0.0.

        Returns:
            Nothing
        """
        ticker = ticker.upper()
        t = yf.Ticker(ticker)
        df = t.history(period="max", auto_adjust=False)
        df.index = pd.to_datetime(df.index)
        self._tickers[ticker] = df
        self._names[ticker] = t.info["longName"] if "longName" in t.info else ""
        self._total_assets[ticker] = t.info["totalAssets"] if "totalAssets" in t.info else None
        self._daily_volume_10_days[ticker] = (
            t.info["averageDailyVolume10Day"] if "averageDailyVolume10Day" in t.info else None
        )
        self._expense_ratio[ticker] = expense_ratio

    def get_market(self) -> Market:  # pragma: no cover
        """Get market data.

        Returns:
            Market data
        """
        sorted_total_assert = [
            item[0]
            for item in sorted(self._total_assets.items(), key=lambda x: x[1] if x[1] is not None else 0, reverse=True)
        ]
        return Market.create_from_ticker(
            {k: self._names[k] for k in sorted_total_assert},
            {k: self._tickers[k] for k in sorted_total_assert},
            {k: self._expense_ratio[k] for k in sorted_total_assert},
        )

    def summary(self) -> pd.DataFrame:
        """Get summary.

        Returns:
            Summary.
        """
        df = pd.DataFrame(
            {
                "established": [self._tickers[t].index[0] for t in self._tickers.keys()],
                "names": [self._names[t] for t in self._tickers.keys()],
                "total_assets": [self._total_assets[t] for t in self._tickers.keys()],
                "volume": [self._daily_volume_10_days[t] for t in self._tickers.keys()],
                "expense_ratio": [self._expense_ratio[t] for t in self._tickers.keys()],
            },
            index=self._tickers.keys(),
        )
        return df.sort_index()

    def since(self, since: datetime) -> Loader:
        """Filter market data.

        Args:
            since: Established date.

        Returns:
            Filtered data.
        """
        tickers = [ticker for ticker in self._tickers.keys() if self._tickers[ticker].index[0] <= since]
        return self.filter(tickers)

    def volume_than(self, volume: int) -> Loader:
        """Filter market data.

        Args:
            volume: Volume for comparing.

        Returns:
            Filtered data.
        """
        tickers = [ticker for ticker in self._tickers.keys() if self._daily_volume_10_days[ticker] >= volume]
        return self.filter(tickers)

    def total_assets_than(self, total: int) -> Loader:
        """Filter market data.

        Args:
            total: Total assets for comparing.

        Returns:
            Filtered data.
        """
        tickers = [ticker for ticker in self._tickers.keys() if self._total_assets[ticker] >= total]
        return self.filter(tickers)

    def filter(self, tickers: List[str]) -> Loader:
        """Filter market data.

        Args:
            since: Established date.

        Returns:
            Filtered data.
        """
        return Loader(
            tickers={key: value for key, value in self._tickers.items() if key in tickers},
            names={key: value for key, value in self._names.items() if key in tickers},
            expense_ratio={key: value for key, value in self._expense_ratio.items() if key in tickers},
            total_assets={key: value for key, value in self._total_assets.items() if key in tickers},
            daily_volume_10_days={key: value for key, value in self._daily_volume_10_days.items() if key in tickers},
        )



"""Class for report result."""
import datetime
import itertools
from typing import List

import numpy as np
import pandas as pd

from mypo.indicator import max_drawdown, max_drawdown_span, yearly_total_return


class Reporter(object):
    """Reporter class of simulation."""

    _index: List[datetime.datetime]
    _weights: List[np.ndarray]
    _tickers: List[List[str]]
    _total_assets: List[np.float64]
    _capital_gain: List[np.float64]
    _income_gain: List[np.float64]
    _cash: List[np.float64]
    _deal: List[np.float64]
    _fee: List[np.float64]
    _capital_gain_tax: List[np.float64]
    _income_gain_tax: List[np.float64]

    def __init__(self) -> None:
        """Construct this object."""
        self._index = []
        self._weights = []
        self._tickers = []
        self._total_assets = []
        self._capital_gain = []
        self._income_gain = []
        self._cash = []
        self._deal = []
        self._fee = []
        self._capital_gain_tax = []
        self._income_gain_tax = []

    def record(
        self,
        at: datetime.datetime,
        weights: np.ndarray,
        tickers: List[str],
        total_assets: np.float64,
        capital_gain: np.float64,
        income_gain: np.float64,
        cash: np.float64,
        deal: np.float64,
        fee: np.float64,
        capital_gain_tax: np.float64,
        income_gain_tax: np.float64,
    ) -> None:
        """Record current situation.

        Args:
            at: Current date.
            weights: Weight of asserts.
            tickers: Tickers.
            total_assets: Current total assets.
            capital_gain: Current capital gain.
            income_gain: Current income gain.
            cash: Current cash.
            deal: Current deal.
            fee: Current fee.
            capital_gain_tax: Current capital gain tax.
            income_gain_tax: Current income gain tax.
        """
        self._index += [at]
        self._weights += [weights]
        self._tickers += [tickers]
        self._total_assets += [total_assets]
        self._capital_gain += [capital_gain]
        self._income_gain += [income_gain]
        self._cash += [cash]
        self._deal += [deal]
        self._fee += [fee]
        self._capital_gain_tax += [capital_gain_tax]
        self._income_gain_tax += [income_gain_tax]

    def history(self) -> pd.DataFrame:
        """Report the result.

        Returns:
            result
        """
        return pd.DataFrame(
            {
                "total_assets": self._total_assets,
                "capital_gain": self._capital_gain,
                "income_gain": self._income_gain,
                "cash": self._cash,
                "deal": self._deal,
                "fee": self._fee,
                "capital_gain_tax": self._capital_gain_tax,
                "income_gain_tax": self._income_gain_tax,
            },
            index=self._index,
        )

    def get_tickers(self) -> List[str]:
        """Get tickers.

        Returns:
            Tickers.
        """
        return list(set(itertools.chain.from_iterable(self._tickers)))

    def history_weights(self) -> pd.DataFrame:
        """Report weights.

        Returns:
            Report of weights.
        """
        tickers = self.get_tickers()
        ret = []
        for i in range(len(self._weights)):
            weight_record = {ticker: 0.0 for ticker in tickers}
            for j in range(len(self._tickers[i])):
                weight_record[self._tickers[i][j]] = self._weights[i][j]
            ret += [weight_record]
        return pd.DataFrame.from_records(ret, index=self._index)

    def summary(self) -> pd.DataFrame:
        """Get summary.

        Returns:
            Summary.
        """
        report = self.history()
        return pd.DataFrame(
            {
                "tickers": [self.get_tickers()],
                "yearly total return": [yearly_total_return(report)],
                "max draw down": [max_drawdown(report)],
                "max draw down span": [max_drawdown_span(report)],
            }
        )



"""Optimizer for weights of portfolio."""

from datetime import datetime
from typing import Optional

import numpy as np
import pandas as pd
from scipy.optimize import minimize

from mypo.common import safe_cast
from mypo.market import Market
from mypo.optimizer.base_optimizer import BaseOptimizer
from mypo.optimizer.objective import covariance


class MeanVarianceOptimizer(BaseOptimizer):
    """Mean variance optimizer."""

    _span: int
    _risk_tolerance: Optional[float]
    _cost_tolerance: Optional[float]

    def __init__(self, span: int = 260, risk_tolerance: float = 0, cost_tolerance: float = 0):
        """Construct this object.

        Args:
            span: Span for evaluation.
            risk_tolerance: Risk tolerance.
            cost_tolerance: Cost tolerance.
        """
        self._span = span
        self._risk_tolerance = risk_tolerance
        self._cost_tolerance = cost_tolerance
        super().__init__([1])

    def optimize(self, market: Market, at: datetime) -> np.float64:
        """Optimize weights.

        Args:
            market: Past market stock prices.
            at: Current date.

        Returns:
            Optimized weights
        """
        historical_data = market.extract(market.get_index() <= pd.to_datetime(at)).get_rate_of_change()
        prices = historical_data.tail(n=self._span).to_numpy()
        Q = covariance(prices)
        R = prices.mean(axis=0)
        n = len(historical_data.columns)
        x = np.ones(n) / n

        def fn(x: np.ndarray) -> np.float64:
            ret: np.float64 = np.dot(np.dot(x, Q), x.T) - self._risk_tolerance * np.dot(R, x)
            return ret

        cons = [{"type": "eq", "fun": lambda x: np.sum(x) - 1}]
        bounds = [[0.0, 1.0] for i in range(n)]

        minout = minimize(fn, x, method="SLSQP", bounds=bounds, constraints=cons, tol=1e-6 * np.max(np.abs(Q)))

        if self._cost_tolerance != 0:
            # Cn = CEo - CEn + TC`|wn - wc| + d(wn - wo)`(wn - wo)
            previous_weights = safe_cast(self._weights)
            optimized_weights = safe_cast(minout.x)
            optimized_value = np.float64(minout.fun)

            def cn(
                x: np.ndarray,
            ) -> np.float64:
                wo_diff = x - optimized_weights
                wp_diff = x - previous_weights
                ret: np.float64 = (
                    optimized_value
                    - fn(x)
                    + self._cost_tolerance * np.dot(wp_diff, wp_diff.T)
                    + np.dot(wo_diff, wo_diff.T)
                )
                return ret

            minout = minimize(
                cn,
                previous_weights,
                method="SLSQP",
                bounds=bounds,
                constraints=cons,
                tol=1e-6 * np.max(np.abs(Q)),
            )

        self._weights = safe_cast(minout.x)
        return np.float64(minout.fun)



# flake8: noqa
__version__ = "0.0.78"  # refresh automatically by CI

from .common import calc_capital_gain_tax, calc_fee, calc_income_gain_tax
from .loader import Loader
from .market import Market, SamplingMethod
from .model_selection import (
    Fold,
    clustering_tickers,
    evaluate_combinations,
    select_by_correlation,
    select_by_regression,
    split_k_folds,
)
from .reporter import Reporter
from .runner import Runner
from .settings import Settings
